import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras.regularizers import l2


def unet_lstm(
    n_classes=1,
    input_shape=(32, 32, 32, 32, 1),
    filters=8,
    activation="tanh",
    reg_val=1e-08,
    drop_val=0.0,
    drop_val_recur=0.0,
    name="unet_lstm",
):
    """unet_lstm -  A model for the spatial and temporal evolution of 3D fields.

    Parameters
    ----------
    n_classes: int, number of classes to classify. For binary applications, use
        a value of 1.
    input_shape: list or tuple of five ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    filter: int, size of the filter for the model. Default filter size
        is set to be 8.
    activation: str or optimizer object, the non-linearity to use. All
        tf.activations are allowed to use. default "tanh".
    reg_val: float, regularization value.
    drop_val: float, dropout value. [0,1]
    drop_val_recur: float, recurrent dropout value [0,1].

    Returns
    ----------
    Model object.
    """

    batch_norm = False
    concat_axis = -1

    inputs = layers.Input(shape=(input_shape))

    x_layer = layers.ConvLSTM3D(
        filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(inputs)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(x_layer)

    conv1 = layers.ConvLSTM3D(
        filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(conv1)

    # x_layer = layers.MaxPooling4D(pool_size=(1, 2, 2 , 2))(conv1) ToDo
    x_layer = layers.ConvLSTM3D(
        2 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(x_layer)

    conv2 = layers.ConvLSTM3D(
        2 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(conv2)

    # x_layer = layers.MaxPooling4D(pool_size=(1, 2, 2, 2))(conv2) ToDo
    x_layer = layers.ConvLSTM3D(
        4 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(x_layer)

    conv3 = layers.ConvLSTM3D(
        4 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(conv3)

    # x_layer = layers.MaxPooling4D(pool_size=(1, 2, 2, 2))(conv3)
    x_layer = layers.ConvLSTM3D(
        8 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(x_layer)

    conv4 = layers.ConvLSTM3D(
        8 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(conv4)

    # x_layer = layers.MaxPooling4D(pool_size=(1, 2, 2 , 2))(conv4) ToDo
    x_layer = layers.ConvLSTM3D(
        16 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(x_layer)

    conv5 = layers.ConvLSTM3D(
        16 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(conv5)

    # x_layer = layers.UpSampling4D(size=(1, 2, 2, 2))(conv5) ToDo
    x_layer = layers.concatenate([x_layer, conv4], axis=concat_axis)

    x_layer = layers.ConvLSTM3D(
        8 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(x_layer)

    x_layer = layers.ConvLSTM3D(
        8 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(x_layer)

    # x_layer = layers.UpSampling4D(size=(1, 2, 2, 2))(x_layer) ToDo
    x_layer = layers.concatenate([x_layer, conv3], axis=concat_axis)

    x_layer = layers.ConvLSTM3D(
        4 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(x_layer)

    x_layer = layers.ConvLSTM3D(
        4 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(x_layer)

    # x_layer = layers.UpSampling4D(size=(1, 2, 2, 2))(x_layer) ToDo
    x_layer = layers.concatenate([x_layer, conv2], axis=concat_axis)

    x_layer = layers.ConvLSTM3D(
        2 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(x_layer)

    x_layer = layers.ConvLSTM3D(
        2 * filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(x_layer)

    # x_layer = layers.UpSampling4D(size=(1, 2, 2))(x_layer) ToDo
    x_layer = layers.concatenate([x_layer, conv1], axis=concat_axis)

    x_layer = layers.ConvLSTM3D(
        filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(x_layer)

    x_layer = layers.ConvLSTM3D(
        filters,
        3,
        activation=activation,
        padding="same",
        kernel_regularizer=l2(reg_val),
        recurrent_regularizer=l2(reg_val),
        bias_regularizer=l2(reg_val),
        dropout=drop_val,
        recurrent_dropout=drop_val_recur,
        return_sequences=True,
    )(x_layer)
    if batch_norm:
        x_layer = layers.BatchNormalization(axis=concat_axis)(x_layer)

    outputs = layers.ConvLSTM3D(
        n_classes, 1, activation="linear", padding="same", return_sequences=False
    )(x_layer)

    return tf.keras.Model(inputs=inputs, outputs=outputs)



"""Model definition for Autoencoder.
"""
import math

from tensorflow.keras import layers, models


def autoencoder(
    input_shape,
    encoding_dim=512,
    n_base_filters=16,
    batchnorm=True,
    batch_size=None,
):
    """Instantiate Autoencoder Architecture.

    Parameters
    ----------
    input_shape: list or tuple of four ints, the shape of the input data. Should be
        scaled to [0,1]. Omit the batch dimension, and include the number of channels.
        Currently, only squares and cubes supported.
    encoding_dim: int, the dimensions of the encoding of the input data. This would
        translate to a latent code of dimensions encoding_dimx1.
    n_base_filters: int, number of base filters the models first convolutional layer.
        The subsequent layers have n_filters which are multiples of n_base_filters.
    batchnorm: bool, whether to use batch normalization in the network.
    batch_size: int, number of samples in each batch. This must be set when training on
        TPUs.
    name: str, name to give to the resulting model object.

    Returns
    -------
    Model object.
    """

    conv_kwds = {"kernel_size": 4, "activation": None, "padding": "same", "strides": 2}

    conv_transpose_kwds = {
        "kernel_size": 4,
        "strides": 2,
        "activation": None,
        "padding": "same",
    }

    dimensions = input_shape[:-1]
    n_dims = len(dimensions)

    if not (n_dims in [2, 3] and dimensions[1:] == dimensions[:-1]):
        raise ValueError("Dimensions should be of square or cube!")

    Conv = getattr(layers, "Conv{}D".format(n_dims))
    ConvTranspose = getattr(layers, "Conv{}DTranspose".format(n_dims))
    n_layers = int(math.log(dimensions[0], 2))

    # Input layer
    inputs = x = layers.Input(shape=input_shape, batch_size=batch_size, name="inputs")

    # Encoder
    for i in range(n_layers):
        n_filters = min(n_base_filters * (2 ** (i)), encoding_dim)

        x = Conv(n_filters, **conv_kwds)(x)
        if batchnorm:
            x = layers.BatchNormalization()(x)
            x = layers.ReLU()(x)

    # Encoding of the input image
    x = layers.Flatten(name="Encoding")(x)

    # Decoder
    x = layers.Reshape((1,) * n_dims + (encoding_dim,))(x)
    for i in range(n_layers)[::-1]:
        n_filters = min(n_base_filters * (2 ** (i)), encoding_dim)

        x = ConvTranspose(n_filters, **conv_transpose_kwds)(x)
        if batchnorm:
            x = layers.BatchNormalization()(x)
            x = layers.LeakyReLU()(x)

    # Output layer
    outputs = Conv(1, 3, activation="sigmoid", padding="same")(x)

    return models.Model(inputs=inputs, outputs=outputs)



import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
import pytest
import scipy.spatial.distance

from .. import metrics


def test_dice():
    x = np.zeros(4)
    y = np.zeros(4)
    out = metrics.dice(x, y, axis=None).numpy()
    assert_allclose(out, 1)

    x = np.ones(4)
    y = np.ones(4)
    out = metrics.dice(x, y, axis=None).numpy()
    assert_allclose(out, 1)

    x = [0.0, 0.0, 1.0, 1.0]
    y = [1.0, 1.0, 1.0, 1.0]
    out = metrics.dice(x, y, axis=None).numpy()
    ref = 1.0 - scipy.spatial.distance.dice(x, y)
    assert_allclose(out, ref)
    jac_out = metrics.jaccard(x, y, axis=None).numpy()
    assert_allclose(out, 2.0 * jac_out / (1.0 + jac_out))

    x = [0.0, 0.0, 1.0, 1.0]
    y = [1.0, 1.0, 0.0, 0.0]
    out = metrics.dice(x, y, axis=None).numpy()
    ref = 1.0 - scipy.spatial.distance.dice(x, y)
    assert_allclose(out, ref, atol=1e-07)
    assert_allclose(out, 0, atol=1e-07)

    x = np.ones((4, 32, 32, 32, 1), dtype=np.float32)
    y = x.copy()
    x[:2, :10, 10:] = 0
    y[:2, :3, 20:] = 0
    y[3:, 10:] = 0
    dices = np.empty(x.shape[0])
    for i in range(x.shape[0]):
        dices[i] = 1.0 - scipy.spatial.distance.dice(x[i].flatten(), y[i].flatten())
    assert_allclose(metrics.dice(x, y, axis=(1, 2, 3, 4)), dices)


def test_generalized_dice():
    shape = (8, 32, 32, 32, 16)
    x = np.zeros(shape)
    y = np.zeros(shape)
    assert_array_equal(metrics.generalized_dice(x, y), np.ones(shape[0]))

    shape = (8, 32, 32, 32, 16)
    x = np.ones(shape)
    y = np.ones(shape)
    assert_array_equal(metrics.generalized_dice(x, y), np.ones(shape[0]))

    shape = (8, 32, 32, 32, 16)
    x = np.ones(shape)
    y = np.zeros(shape)
    # Why aren't the scores exactly zero? Could it be the propagation of floating
    # point inaccuracies when summing?
    assert_allclose(metrics.generalized_dice(x, y), np.zeros(shape[0]), atol=1e-03)

    x = np.ones((4, 32, 32, 32, 1), dtype=np.float64)
    y = x.copy()
    x[:2, :10, 10:] = 0
    y[:2, :3, 20:] = 0
    y[3:, 10:] = 0
    # Dice is similar to generalized Dice for one class. The weight factor
    # makes the generalized form slightly different from Dice.
    gd = metrics.generalized_dice(x, y, axis=(1, 2, 3)).numpy()
    dd = metrics.dice(x, y, axis=(1, 2, 3, 4)).numpy()
    assert_allclose(gd, dd, rtol=1e-02)  # is this close enough?


def test_jaccard():
    x = np.zeros(4)
    y = np.zeros(4)
    out = metrics.jaccard(x, y, axis=None).numpy()
    assert_allclose(out, 1)

    x = np.ones(4)
    y = np.ones(4)
    out = metrics.jaccard(x, y, axis=None).numpy()
    assert_allclose(out, 1)

    x = [0.0, 0.0, 1.0, 1.0]
    y = [1.0, 1.0, 1.0, 1.0]
    out = metrics.jaccard(x, y, axis=None).numpy()
    ref = 1.0 - scipy.spatial.distance.jaccard(x, y)
    assert_allclose(out, ref)
    dice_out = metrics.dice(x, y, axis=None).numpy()
    assert_allclose(out, dice_out / (2.0 - dice_out))

    x = [0.0, 0.0, 1.0, 1.0]
    y = [1.0, 1.0, 0.0, 0.0]
    out = metrics.jaccard(x, y, axis=None).numpy()
    ref = 1.0 - scipy.spatial.distance.jaccard(x, y)
    assert_allclose(out, ref, atol=1e-07)
    assert_allclose(out, 0, atol=1e-07)

    x = np.ones((4, 32, 32, 32, 1), dtype=np.float32)
    y = x.copy()
    x[:2, :10, 10:] = 0
    y[:2, :3, 20:] = 0
    y[3:, 10:] = 0
    jaccards = np.empty(x.shape[0])
    for i in range(x.shape[0]):
        jaccards[i] = 1.0 - scipy.spatial.distance.jaccard(
            x[i].flatten(), y[i].flatten()
        )
    assert_allclose(metrics.jaccard(x, y, axis=(1, 2, 3, 4)), jaccards)


def test_tversky():
    shape = (4, 32, 32, 32, 1)
    y_pred = np.random.rand(*shape).astype(np.float64)
    y_true = np.random.randint(2, size=shape).astype(np.float64)

    # Test that tversky and dice are same when alpha = beta = 0.5
    dice = metrics.dice(y_true, y_pred).numpy()
    tversky = metrics.tversky(
        y_true, y_pred, axis=(1, 2, 3), alpha=0.5, beta=0.5
    ).numpy()
    assert_allclose(dice, tversky)

    # Test that tversky and jaccard are same when alpha = beta = 1.0
    jaccard = metrics.jaccard(y_true, y_pred).numpy()
    tversky = metrics.tversky(
        y_true, y_pred, axis=(1, 2, 3), alpha=1.0, beta=1.0
    ).numpy()
    assert_allclose(jaccard, tversky)

    with pytest.raises(ValueError):
        metrics.tversky([0.0, 0.0, 1.0], [1.0, 0.0, 1.0], axis=0)



"""Tests for `nobrainer.processing.checkpoint`."""

import os

import numpy as np
from numpy.testing import assert_allclose
import tensorflow as tf

from nobrainer.dataset import Dataset
from nobrainer.models import meshnet
from nobrainer.processing.segmentation import Segmentation


def _get_toy_dataset():
    data_shape = (8, 8, 8, 8, 1)
    train = tf.data.Dataset.from_tensors(
        (np.random.rand(*data_shape), np.random.randint(0, 1, data_shape))
    )
    return Dataset(train, data_shape[0], data_shape[1:4], 1)


def _assert_model_weights_allclose(model1, model2):
    for layer1, layer2 in zip(model1.model.layers, model2.model.layers):
        weights1 = layer1.get_weights()
        weights2 = layer2.get_weights()
        assert len(weights1) == len(weights2)
        for index in range(len(weights1)):
            assert_allclose(weights1[index], weights2[index], rtol=1e-06, atol=1e-08)


def test_checkpoint(tmp_path):
    train = _get_toy_dataset()

    checkpoint_filepath = os.path.join(tmp_path, "checkpoint-epoch_{epoch:03d}")
    model1 = Segmentation.init_with_checkpoints(
        meshnet,
        checkpoint_filepath=checkpoint_filepath,
    )
    model1.fit(
        dataset_train=train,
        epochs=2,
    )

    model2 = Segmentation.init_with_checkpoints(
        meshnet,
        checkpoint_filepath=checkpoint_filepath,
    )
    _assert_model_weights_allclose(model1, model2)
    model2.fit(
        dataset_train=train,
        epochs=3,
    )

    model3 = Segmentation.init_with_checkpoints(
        meshnet,
        checkpoint_filepath=checkpoint_filepath,
    )
    _assert_model_weights_allclose(model2, model3)


def test_warm_start_workflow(tmp_path):
    train = _get_toy_dataset()

    checkpoint_dir = os.path.join(tmp_path, "checkpoints")
    checkpoint_filepath = os.path.join(checkpoint_dir, "{epoch:03d}")
    if not os.path.exists(checkpoint_dir):
        os.mkdir(checkpoint_dir)

    for iteration in range(2):
        bem = Segmentation.init_with_checkpoints(
            meshnet,
            checkpoint_filepath=checkpoint_filepath,
        )
        if iteration == 0:
            assert bem.model is None
        else:
            assert bem.model is not None
            for layer in bem.model.layers:
                for weight_array in layer.get_weights():
                    assert np.count_nonzero(weight_array)
        bem.fit(
            dataset_train=train,
            epochs=2,
        )



import tensorflow as tf


def average_volume_difference():
    raise NotImplementedError()


def dice(y_true, y_pred, axis=(1, 2, 3, 4)):
    """Calculate Dice similarity between labels and predictions.
    Dice similarity is in [0, 1], where 1 is perfect overlap and 0 is no
    overlap. If both labels and predictions are empty (e.g., all background),
    then Dice similarity is 1.
    If we assume the inputs are rank 5 [`(batch, x, y, z, classes)`], then an
    axis parameter of `(1, 2, 3)` will result in a tensor that contains a Dice
    score for every class in every item in the batch. The shape of this tensor
    will be `(batch, classes)`. If the inputs only have one class (e.g., binary
    segmentation), then an axis parameter of `(1, 2, 3, 4)` should be used.
    This will result in a tensor of shape `(batch,)`, where every value is the
    Dice similarity for that prediction.
    Implemented according to https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4533825/#Equ6
    Returns
    -------
    Tensor of Dice similarities.
    Citations
    ---------
    Taha AA, Hanbury A. Metrics for evaluating 3D medical image segmentation:
        analysis, selection, and tool. BMC Med Imaging. 2015;15:29. Published 2015
        Aug 12. doi:10.1186/s12880-015-0068-x
    """
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.cast(y_true, y_pred.dtype)
    eps = tf.keras.backend.epsilon()

    intersection = tf.reduce_sum(y_true * y_pred, axis=axis)
    summation = tf.reduce_sum(y_true, axis=axis) + tf.reduce_sum(y_pred, axis=axis)
    return (2 * intersection + eps) / (summation + eps)


def generalized_dice(y_true, y_pred, axis=(1, 2, 3)):
    """Calculate Generalized Dice similarity. This is useful for multi-class
    predictions.
    If we assume the inputs are rank 5 [`(batch, x, y, z, classes)`], then an
    axis parameter of `(1, 2, 3)` should be used. This will result in a tensor
    of shape `(batch,)`, where every value is the Generalized Dice similarity
    for that prediction, across all classes.
    Returns
    -------
    Tensor of Generalized Dice similarities.
    """
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.cast(y_true, y_pred.dtype)

    if y_true.get_shape().ndims < 2 or y_pred.get_shape().ndims < 2:
        raise ValueError("y_true and y_pred must be at least rank 2.")

    epsilon = tf.keras.backend.epsilon()

    w = tf.math.reciprocal(tf.square(tf.reduce_sum(y_true, axis=axis)))
    w = tf.where(tf.math.is_finite(w), w, epsilon)
    num = 2 * tf.reduce_sum(w * tf.reduce_sum(y_true * y_pred, axis=axis), axis=-1)
    den = tf.reduce_sum(w * tf.reduce_sum(y_true + y_pred, axis=axis), axis=-1)
    gdice = (num + epsilon) / (den + epsilon)
    return gdice


def hamming(y_true, y_pred, axis=(1, 2, 3)):
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.cast(y_true, y_pred.dtype)
    return tf.reduce_mean(tf.not_equal(y_pred, y_true), axis=axis)


def haussdorf():
    raise NotADirectoryError()


def jaccard(y_true, y_pred, axis=(1, 2, 3, 4)):
    """Calculate Jaccard similarity between labels and predictions.
    Jaccard similarity is in [0, 1], where 1 is perfect overlap and 0 is no
    overlap. If both labels and predictions are empty (e.g., all background),
    then Jaccard similarity is 1.
    If we assume the inputs are rank 5 [`(batch, x, y, z, classes)`], then an
    axis parameter of `(1, 2, 3)` will result in a tensor that contains a Jaccard
    score for every class in every item in the batch. The shape of this tensor
    will be `(batch, classes)`. If the inputs only have one class (e.g., binary
    segmentation), then an axis parameter of `(1, 2, 3, 4)` should be used.
    This will result in a tensor of shape `(batch,)`, where every value is the
    Jaccard similarity for that prediction.
    Implemented according to https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4533825/#Equ7
    Returns
    -------
    Tensor of Jaccard similarities.
    Citations
    ---------
    Taha AA, Hanbury A. Metrics for evaluating 3D medical image segmentation:
        analysis, selection, and tool. BMC Med Imaging. 2015;15:29. Published 2015
        Aug 12. doi:10.1186/s12880-015-0068-x
    """
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.cast(y_true, y_pred.dtype)
    eps = tf.keras.backend.epsilon()

    intersection = tf.reduce_sum(y_true * y_pred, axis=axis)
    union = tf.reduce_sum(y_true, axis=axis) + tf.reduce_sum(y_pred, axis=axis)
    return (intersection + eps) / (union - intersection + eps)


def tversky(y_true, y_pred, axis=(1, 2, 3), alpha=0.3, beta=0.7):
    y_pred = tf.convert_to_tensor(y_pred)
    y_true = tf.cast(y_true, y_pred.dtype)

    if y_true.get_shape().ndims < 2 or y_pred.get_shape().ndims < 2:
        raise ValueError("y_true and y_pred must be at least rank 2.")

    eps = tf.keras.backend.epsilon()

    num = tf.reduce_sum(y_pred * y_true, axis=axis)
    den = (
        num
        + alpha * tf.reduce_sum(y_pred * (1 - y_true), axis=axis)
        + beta * tf.reduce_sum((1 - y_pred) * y_true, axis=axis)
    )
    # Sum over classes.
    return tf.reduce_sum((num + eps) / (den + eps), axis=-1)



"""Base classes for all estimators."""

import inspect
import os
from pathlib import Path
import pickle as pk

import tensorflow as tf


def get_strategy(multi_gpu):
    if multi_gpu:
        return tf.distribute.MirroredStrategy()
    return tf.distribute.get_strategy()


class BaseEstimator:
    """Base class for all high-level models in Nobrainer."""

    state_variables = []
    model_ = None

    def __init__(self, checkpoint_filepath=None, multi_gpu=True):
        self.checkpoint_tracker = None
        if checkpoint_filepath:
            from .checkpoint import CheckpointTracker

            self.checkpoint_tracker = CheckpointTracker(self, checkpoint_filepath)

        self.strategy = get_strategy(multi_gpu)

    @property
    def model(self):
        return self.model_

    def save(self, save_dir):
        """Saves a trained model"""
        if self.model_ is None:
            raise ValueError("Model is undefined. Please train or load a model")
        self.model_.save(save_dir)
        model_info = {"classname": self.__class__.__name__, "__init__": {}}
        for key in inspect.signature(self.__init__).parameters:
            # TODO this assumes that all parameters passed to __init__
            # are stored as members, which doesn't leave room for
            # parameters that are specific to the runtime context.
            # (e.g. multi_gpu).
            if key == "multi_gpu" or key == "checkpoint_filepath":
                continue
            model_info["__init__"][key] = getattr(self, key)
        for val in self.state_variables:
            model_info[val] = getattr(self, val)
        model_file = Path(save_dir) / "model_params.pkl"
        with open(model_file, "wb") as fp:
            pk.dump(model_info, fp)

    @classmethod
    def load(
        cls,
        model_dir,
        multi_gpu=True,
        custom_objects=None,
        compile=False,
    ):
        """Loads a trained model from a save directory"""
        model_dir = Path(str(model_dir).rstrip(os.pathsep))
        assert model_dir.exists() and model_dir.is_dir()
        model_file = model_dir / "model_params.pkl"
        with open(model_file, "rb") as fp:
            model_info = pk.load(fp)
        if model_info["classname"] != cls.__name__:
            raise ValueError(f"Model class does not match {cls.__name__}")
        del model_info["classname"]

        klass = cls(**model_info["__init__"])
        del model_info["__init__"]
        for key, value in model_info.items():
            setattr(klass, key, value)

        klass.strategy = get_strategy(multi_gpu)
        with klass.strategy.scope():
            klass.model_ = tf.keras.models.load_model(
                model_dir,
                custom_objects=custom_objects,
                compile=compile,
            )
        return klass

    @classmethod
    def init_with_checkpoints(
        cls,
        model_name,
        checkpoint_filepath,
        multi_gpu=True,
        custom_objects=None,
        compile=False,
        model_args=None,
    ):
        """Initialize a model for training, either from the latest
        checkpoint found, or from scratch if no checkpoints are
        found. This is useful for long-running model fits that may be
        interrupted or preepmted during training and need to pick up
        where they left off.

        model_name: str or Module in nobrainer.models, the base model
        for this estimator.

        checkpoint_filepath: str, path to which checkpoints will be
        saved and loaded. Supports the epoch and block flormating
        parameters supported by tensorflows ModelCheckpoint,
        e.g. <path_to_checkpoint_dir>/{epoch:03d}

        """
        from .checkpoint import CheckpointTracker

        checkpoint_tracker = CheckpointTracker(cls, checkpoint_filepath)
        estimator = checkpoint_tracker.load(
            multi_gpu=multi_gpu,
            custom_objects=custom_objects,
            compile=compile,
        )
        if not estimator:
            estimator = cls(model_name, model_args=model_args)
        estimator.checkpoint_tracker = checkpoint_tracker
        checkpoint_tracker.estimator = estimator
        return estimator


class TransformerMixin:
    """Mixin class for all transformers in scikit-learn."""

    def fit_transform(self, X, y=None, **fit_params):
        """
        Fit to data, then transform it.
        Fits transformer to `X` and `y` with optional parameters `fit_params`
        and returns a transformed version of `X`.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Input samples.
        y :  array-like of shape (n_samples,) or (n_samples, n_outputs), \
                default=None
            Target values (None for unsupervised transformations).
        **fit_params : dict
            Additional fit parameters.
        Returns
        -------
        X_new : ndarray array of shape (n_samples, n_features_new)
            Transformed array.
        """
        # non-optimized default implementation; override when a better
        # method is possible for a given clustering algorithm
        if y is None:
            # fit method of arity 1 (unsupervised transformation)
            return self.fit(X, **fit_params).transform(X)
        else:
            # fit method of arity 2 (supervised transformation)
            return self.fit(X, y, **fit_params).transform(X)



"""Implementations of Bayesian neural networks."""

import tensorflow as tf
import tensorflow_probability as tfp

from ..bayesian_utils import divergence_fn_bayesian, prior_fn_for_bayesian
from ..layers.dropout import BernoulliDropout, ConcreteDropout

tfk = tf.keras
tfkl = tfk.layers
tfpl = tfp.layers
tfd = tfp.distributions


def variational_meshnet(
    n_classes,
    input_shape,
    receptive_field=67,
    filters=71,
    no_examples=3000,
    is_monte_carlo=False,
    dropout=None,
    activation=tf.nn.relu,
    batch_size=None,
    name="variational_meshnet",
):
    """Instantiate variational MeshNet model.

    Please see https://arxiv.org/abs/1805.10863 for Meshnet related information.

    Parameters
    ----------
    n_classes: int, number of classes to classify. For binary applications, use
        a value of 1.
    input_shape: list or tuple of four ints, the shape of the input data. Omit
        the batch dimension, and include the number of channels.
    receptive_field: {37, 67, 129}, the receptive field of the model. According
        to the MeshNet manuscript, the receptive field should be similar to your
        input shape. The actual receptive field is the cube of the value provided.
    filters: int, number of filters per volumetric convolution. The original
        MeshNet manuscript uses 21 filters for a binary segmentation task
        (i.e., brain extraction) and 71 filters for a multi-class segmentation task.
    activation: str or optimizer object, the non-linearity to use.
    scale_factor: A tf float 32 variable to scale up the KLD loss.
    is_monte_carlo: bool, only Related to dropout version!
    dropout: string, type of dropout layer.
    batch_size: int, number of samples in each batch. This must be set when
        training on TPUs.
    name: str, name to give to the resulting model object.
    Set priors, divergence and posteriors for training with this model.

    Returns
    -------
    Model object.

    Raises
    ------
    ValueError if receptive field is not an allowable value.
    """

    if receptive_field not in {37, 67, 129}:
        raise ValueError("unknown receptive field. Legal values are 37, 67, and 129.")

    def one_layer(x, layer_num, no_examples=3000, dilation_rate=(1, 1, 1)):
        x = tfpl.Convolution3DFlipout(
            filters,
            kernel_size=3,
            padding="same",
            dilation_rate=dilation_rate,
            kernel_prior_fn=prior_fn_for_bayesian(),
            kernel_divergence_fn=divergence_fn_bayesian(
                prior_std=1.0, examples_per_epoch=no_examples
            ),
            name="layer{}/vwnconv3d".format(layer_num),
        )(x)
        if dropout is None:
            pass
        elif dropout == "bernoulli":
            x = BernoulliDropout(
                rate=0.5,
                is_monte_carlo=is_monte_carlo,
                scale_during_training=False,
                name="layer{}/bernoulli_dropout".format(layer_num),
            )(x)
        elif dropout == "concrete":
            x = ConcreteDropout(
                is_monte_carlo=is_monte_carlo,
                temperature=0.02,
                use_expectation=is_monte_carlo,
                name="layer{}/concrete_dropout".format(layer_num),
            )(x)
        else:
            raise ValueError("unknown dropout layer, {}".format(dropout))
        x = tfkl.Activation(activation, name="layer{}/activation".format(layer_num))(x)
        return x

    inputs = tfkl.Input(shape=input_shape, batch_size=batch_size, name="inputs")

    if receptive_field == 37:
        x = one_layer(inputs, 1)
        x = one_layer(x, 2)
        x = one_layer(x, 3)
        x = one_layer(x, 4, dilation_rate=(2, 2, 2))
        x = one_layer(x, 5, dilation_rate=(4, 4, 4))
        x = one_layer(x, 6, dilation_rate=(8, 8, 8))
        x = one_layer(x, 7)
    elif receptive_field == 67:
        x = one_layer(inputs, 1)
        x = one_layer(x, 2)
        x = one_layer(x, 3, dilation_rate=(2, 2, 2))
        x = one_layer(x, 4, dilation_rate=(4, 4, 4))
        x = one_layer(x, 5, dilation_rate=(8, 8, 8))
        x = one_layer(x, 6, dilation_rate=(16, 16, 16))
        x = one_layer(x, 7)
    elif receptive_field == 129:
        x = one_layer(inputs, 1)
        x = one_layer(x, 2, dilation_rate=(2, 2, 2))
        x = one_layer(x, 3, dilation_rate=(4, 4, 4))
        x = one_layer(x, 4, dilation_rate=(8, 8, 8))
        x = one_layer(x, 5, dilation_rate=(16, 16, 16))
        x = one_layer(x, 6, dilation_rate=(32, 32, 32))
        x = one_layer(x, 7)

    x = tfpl.Convolution3DFlipout(
        filters=n_classes,
        kernel_size=1,
        padding="same",
        name="classification/vwnconv3d",
    )(x)

    final_activation = "sigmoid" if n_classes == 1 else "softmax"
    x = tfkl.Activation(final_activation, name="segmentation/activation")(x)

    return tf.keras.Model(inputs=inputs, outputs=x, name=name)



"""Checkpointing utils"""

from glob import glob
import logging
import os

import tensorflow as tf


class CheckpointTracker(tf.keras.callbacks.ModelCheckpoint):
    """Class for saving/loading estimators at/from checkpoints."""

    def __init__(self, estimator, file_path, **kwargs):
        """
        estimator: BaseEstimator, instance of an estimator (e.g., Segmentation).
        file_path: str, directory to/from which to save or load.
        """
        self.estimator = estimator
        super().__init__(file_path, **kwargs)

    def _save_model(self, epoch, batch, logs):
        """Save the current state of the estimator. This overrides the
        base class implementation to save `nobrainer` specific info.

        epoch: int, the index of the epoch that just finished.
        batch: int, the index of the batch that just finished.
        logs: dict, logging info passed into on_epoch_end or on_batch_end.
        """
        self.save(self._get_file_path(epoch, batch, logs))

    def save(self, directory):
        """Save the current state of the estimator.
        directory: str, path in which to save the model.
        """
        logging.info(f"Saving to dir {directory}")
        self.estimator.save(directory)

    def load(
        self,
        multi_gpu=True,
        custom_objects=None,
        compile=False,
        model_args=None,
    ):
        """Loads the most-recently created checkpoint from the
        checkpoint directory.
        """
        checkpoints = glob(os.path.join(os.path.dirname(self.filepath), "*/"))
        if not checkpoints:
            self.last_epoch = 0
            return None

        # TODO, we should probably exclude non-checkpoint files here,
        # and maybe parse the filename for the epoch number
        self.last_epoch = len(checkpoints)

        latest = max(checkpoints, key=os.path.getctime)
        self.estimator = self.estimator.load(
            latest,
            multi_gpu=multi_gpu,
            custom_objects=custom_objects,
            compile=compile,
        )
        logging.info(f"Loaded estimator from {latest}.")
        return self.estimator



"""Model definition for Attention U-Net.
Adapted from https://github.com/nikhilroxtomar/Semantic-Segmentation-Architecture/blob/main/TensorFlow/attention-unet.py
"""  # noqa: E501

from tensorflow.keras import layers
import tensorflow.keras.layers as L
from tensorflow.keras.models import Model


def conv_block(x, num_filters):
    x = L.Conv3D(num_filters, 3, padding="same")(x)
    x = L.BatchNormalization()(x)
    x = L.Activation("relu")(x)

    x = L.Conv3D(num_filters, 3, padding="same")(x)
    x = L.BatchNormalization()(x)
    x = L.Activation("relu")(x)

    return x


def encoder_block(x, num_filters):
    x = conv_block(x, num_filters)
    p = L.MaxPool3D()(x)
    return x, p


def attention_gate(g, s, num_filters):
    Wg = L.Conv3D(num_filters, 1, padding="same")(g)
    Wg = L.BatchNormalization()(Wg)

    Ws = L.Conv3D(num_filters, 1, padding="same")(s)
    Ws = L.BatchNormalization()(Ws)

    out = L.Activation("relu")(Wg + Ws)
    out = L.Conv3D(num_filters, 1, padding="same")(out)
    out = L.Activation("sigmoid")(out)

    return out * s


def decoder_block(x, s, num_filters):
    x = L.UpSampling3D()(x)
    s = attention_gate(x, s, num_filters)
    x = L.Concatenate()([x, s])
    x = conv_block(x, num_filters)
    return x


def attention_unet(n_classes, input_shape):
    """Inputs"""
    inputs = L.Input(input_shape)

    """ Encoder """
    s1, p1 = encoder_block(inputs, 64)
    s2, p2 = encoder_block(p1, 128)
    s3, p3 = encoder_block(p2, 256)

    b1 = conv_block(p3, 512)

    """ Decoder """
    d1 = decoder_block(b1, s3, 256)
    d2 = decoder_block(d1, s2, 128)
    d3 = decoder_block(d2, s1, 64)

    """ Outputs """
    outputs = L.Conv3D(n_classes, 1, padding="same")(d3)

    final_activation = "sigmoid" if n_classes == 1 else "softmax"
    outputs = layers.Activation(final_activation)(outputs)

    """ Model """
    return Model(inputs=inputs, outputs=outputs, name="Attention_U-Net")


if __name__ == "__main__":
    n_classes = 50
    input_shape = (256, 256, 256, 3)
    model = attention_unet(n_classes, input_shape)
    model.summary()



"""Model definition for HighResNet.
"""

import tensorflow as tf
from tensorflow.keras import layers

from ..layers.padding import ZeroPadding3DChannels


def highresnet(
    n_classes, input_shape, activation="relu", dropout_rate=0, name="highresnet"
):
    """
    Instantiate a 3D HighResnet Architecture.
    Implementation is according to the
    https://arxiv.org/abs/1707.01992
    Args:
    n_classes(int): number of classes
    input_shape(tuple):four ints representing the shape of 3D input
    activation(str): all tf.keras.activations are allowed
    dropout_rate(int): [0,1].
    """

    conv_kwds = {"kernel_size": (3, 3, 3), "padding": "same"}

    n_base_filters = 16

    inputs = layers.Input(shape=input_shape)
    x = layers.Conv3D(n_base_filters, **conv_kwds)(inputs)

    for ii in range(3):
        skip = x
        x = layers.BatchNormalization()(x)
        x = layers.Activation(activation)(x)
        x = layers.Conv3D(n_base_filters, **conv_kwds)(x)
        x = layers.BatchNormalization()(x)
        x = layers.Activation(activation)(x)
        x = layers.Conv3D(n_base_filters, **conv_kwds)(x)
        x = layers.Add()([x, skip])

    x = ZeroPadding3DChannels(8)(x)
    for ii in range(3):
        skip = x
        x = layers.BatchNormalization()(x)
        x = layers.Activation(activation)(x)
        x = layers.Conv3D(n_base_filters * 2, dilation_rate=2, **conv_kwds)(x)
        x = layers.BatchNormalization()(x)
        x = layers.Activation(activation)(x)
        x = layers.Conv3D(n_base_filters * 2, dilation_rate=2, **conv_kwds)(x)
        x = layers.Add()([x, skip])

    x = ZeroPadding3DChannels(16)(x)
    for ii in range(3):
        skip = x
        x = layers.BatchNormalization()(x)
        x = layers.Activation(activation)(x)
        x = layers.Conv3D(n_base_filters * 4, dilation_rate=4, **conv_kwds)(x)
        x = layers.BatchNormalization()(x)
        x = layers.Activation(activation)(x)
        x = layers.Conv3D(n_base_filters * 4, dilation_rate=4, **conv_kwds)(x)
        x = layers.Add()([x, skip])

    x = layers.Conv3D(filters=n_classes, kernel_size=(1, 1, 1), padding="same")(x)

    final_activation = "sigmoid" if n_classes == 1 else "softmax"
    x = layers.Activation(final_activation)(x)

    # QUESTION: where should dropout go?

    return tf.keras.Model(inputs=inputs, outputs=x, name=name)



"""Implementations of loss functions for 3D semantic segmentation."""

import tensorflow as tf
from tensorflow.python.keras.losses import LossFunctionWrapper
from tensorflow.python.keras.utils.losses_utils import ReductionV2

from . import metrics


def dice(y_true, y_pred, axis=(1, 2, 3, 4)):
    return 1.0 - metrics.dice(y_true=y_true, y_pred=y_pred, axis=axis)


class Dice(LossFunctionWrapper):
    """Computes one minus the Dice similarity between labels and predictions.
    For example, if `y_true` is [0., 0., 1., 1.] and `y_pred` is [1., 1., 1., 0.]
    then the Dice loss is 0.6. The Dice similarity between these tensors is 0.4.

    Use this loss only for binary semantic segmentation tasks. The default value
    for the axis parameter is meant for models which output a shape of
    `(batch, x, y, z, 1)`. Values in `y_true` and `y_pred` should be in the
    range [0, 1].

    Usage:

    ```python
    dice = nobrainer.losses.Dice(axis=None)
    loss = dice([0., 0., 1., 1.], [1., 1., 1., 0.])
    print('Loss: ', loss.numpy())  # Loss: 0.6
    ```

    Usage with tf.keras API:
    ```python
    model = tf.keras.Model(inputs, outputs)
    model.compile('sgd', loss=nobrainer.losses.Dice())
    ```
    """

    def __init__(self, axis=(1, 2, 3, 4), reduction=ReductionV2.AUTO, name="dice"):
        super().__init__(dice, axis=axis, reduction=reduction, name=name)


def focal_tversky(y_true, y_pred):
    """
    https://arxiv.org/pdf/1810.07842.pdf
    """
    raise NotImplementedError()


def generalized_dice(y_true, y_pred, axis=(1, 2, 3)):
    return 1.0 - metrics.generalized_dice(y_true=y_true, y_pred=y_pred, axis=axis)


class GeneralizedDice(LossFunctionWrapper):
    """Computes one minus the generalized Dice similarity between labels and
    predictions. For example, if `y_true` is [[0., 0., 1., 1.]] and `y_pred` is
    [[1., 1., 1., 0.]] then the generalized Dice loss is 0.60. The generalized
    Dice similarity between these tensors is 0.40.

    Use this loss for binary or multi-class semantic segmentation tasks. The
    default value for the axis parameter is meant for models which output a
    shape of `(batch, x, y, z, n_classes)`. Values in `y_true` and `y_pred`
    should be in the range [0, 1].

    Usage:

    ```python
    generalized_dice = nobrainer.losses.GeneralizedDice(axis=1)
    generalized_dice([[0., 0., 1., 1.]], [[1., 1., 1., 0.]])
    print('Loss: ', loss.numpy())  # Loss: 0.60
    ```

    Usage with tf.keras API:
    ```python
    model = tf.keras.Model(inputs, outputs)
    model.compile('sgd', loss=nobrainer.losses.GeneralizedDice())
    ```
    """

    def __init__(
        self, axis=(1, 2, 3), reduction=ReductionV2.AUTO, name="generalized_dice"
    ):
        super().__init__(generalized_dice, axis=axis, reduction=reduction, name=name)


def jaccard(y_true, y_pred, axis=(1, 2, 3, 4)):
    return 1.0 - metrics.jaccard(y_true=y_true, y_pred=y_pred, axis=axis)


class Jaccard(LossFunctionWrapper):
    """Computes one minus the Jaccard similarity between labels and predictions.
    For example, if `y_true` is [0., 0., 1., 1.] and `y_pred` is [1., 1., 1., 0.]
    then the Jaccard loss is 0.75. The Jaccard similarity between these tensors
    is 0.25.

    Use this loss only for binary semantic segmentation tasks. The default value
    for the axis parameter is meant for models which output a shape of
    `(batch, x, y, z, 1)`. Values in `y_true` and `y_pred` should be in the
    range [0, 1].

    Usage:

    ```python
    jaccard = nobrainer.losses.Jaccard(axis=None)
    loss = jaccard([0., 0., 1., 1.], [1., 1., 1., 0.])
    print('Loss: ', loss.numpy())  # Loss: 0.75
    ```

    Usage with tf.keras API:
    ```python
    model = tf.keras.Model(inputs, outputs)
    model.compile('sgd', loss=nobrainer.losses.Jaccard())
    ```
    """

    def __init__(self, axis=(1, 2, 3, 4), reduction=ReductionV2.AUTO, name="jaccard"):
        super().__init__(jaccard, axis=axis, reduction=reduction, name=name)


def tversky(y_true, y_pred, axis=(1, 2, 3), alpha=0.3, beta=0.7):
    n_classes = tf.cast(tf.shape(y_pred)[-1], tf.float32)
    # Scores are in dynamic range of [0, n_classes].
    scores = metrics.tversky(
        y_true=y_true, y_pred=y_pred, axis=axis, alpha=alpha, beta=beta
    )
    return 1.0 - (scores / n_classes)


class Tversky(LossFunctionWrapper):
    """Computes Tversky loss between labels and predictions.

    Use this loss for binary or multi-class semantic segmentation tasks. The
    default value for the axis parameter is meant for binary or multi-class
    predictions. Values in `y_true` and `y_pred` should be in the range [0, 1].
    The default values for alpha and beta are set according to recommendations
    in the Tverky loss manuscript.

    Usage with tf.keras API:
    ```python
    model = tf.keras.Model(inputs, outputs)
    model.compile('sgd', loss=nobrainer.losses.Tversky())
    ```
    """

    def __init__(
        self,
        axis=(1, 2, 3),
        alpha=0.3,
        beta=0.7,
        reduction=ReductionV2.AUTO,
        name="tversky",
    ):
        super().__init__(
            tversky, axis=axis, alpha=alpha, beta=beta, reduction=reduction, name=name
        )


def elbo(y_true, y_pred, model, num_examples, from_logits=False):
    """Labels should be integers in `[0, n)`."""
    scc_fn = tf.losses.SparseCategoricalCrossentropy(from_logits=from_logits)
    neg_log_likelihood = scc_fn(y_true, y_pred)
    kl = sum(model.losses) / num_examples
    elbo_loss = neg_log_likelihood + kl
    return elbo_loss


class ELBO(LossFunctionWrapper):
    """Loss to minimize Evidence Lower Bound (ELBO).

    Use this loss for multiclass variational segmentation.
    Labels should not be one-hot encoded.
    """

    def __init__(
        self,
        model,
        num_examples,
        from_logits=False,
        reduction=ReductionV2.AUTO,
        name="ELBO",
    ):
        super().__init__(
            elbo,
            model=model,
            num_examples=num_examples,
            from_logits=from_logits,
            name=name,
            reduction=reduction,
        )


def wasserstein(y_true, y_pred):
    return y_true * y_pred


class Wasserstein(LossFunctionWrapper):
    """Computes Wasserstein loss between labels and predictions.

    Aims to score the realness or fakeness of a given image and measures the Earth Mover (EM)
    distance. Use this loss for training GANs. Values for `y_true` is 1 or -1 for real and fake,
    and `y_pred` is from discriminator. Use in combination with gradient clipping or gradient
    penalty (WassersteinGP defined below).
    https://arxiv.org/abs/1701.07875

    Usage:
    ```python
    real_pred = discriminator(real)
    fake_pred = discriminator(fake)

    wasserstein_loss = Wasserstein()
    disciminator_loss = wasserstein_loss(1, real_pred) + wasserstein_loss(-1, fake_pred)
    generator_loss = wasserstein_loss(1, fake_pred)
    ```
    """

    def __init__(self, reduction=ReductionV2.AUTO, name="wasserstein"):
        super().__init__(wasserstein, reduction=reduction, name=name)


def gradient_penalty(gradients, real_pred, gp_weight=10, epsilon_weight=0.001):
    gradients_squared = tf.square(gradients)
    gradients_sqr_sum = tf.reduce_sum(
        gradients_squared, axis=tf.range(1, tf.rank(gradients_squared))
    )
    gradient_l2_norm = tf.sqrt(gradients_sqr_sum)

    gradient_penalty = gp_weight * tf.square(1 - gradient_l2_norm)

    epsilon_loss = epsilon_weight * tf.square(real_pred)

    return gradient_penalty + epsilon_loss


class GradientPenalty(LossFunctionWrapper):
    """Computes Gradient Penalty for Wasserstein Loss.
    Improvement to gradient clipping for Wasserstein-GAN to enforce the Lipschitz constraint.
    Uses the points interpolated between real and fake to ensure a gradient norm of 1.
    https://arxiv.org/pdf/1704.00028.pdf

    Usage:
    ```python
    fakes = generator(latents)

    wasserstein_gp = WassersteinGP(discriminator=discriminator, alpha=alpha)
    gradient_penalty = wasserstein_gp(reals, fakes)
    ```
    """

    def __init__(
        self,
        gp_weight=10,
        epsilon_weight=0.001,
        reduction=ReductionV2.AUTO,
        name="wasserstein_gp",
    ):
        super().__init__(
            gradient_penalty,
            gp_weight=gp_weight,
            epsilon_weight=epsilon_weight,
            reduction=reduction,
            name=name,
        )


def get(loss):
    """Wrapper for `tf.keras.losses.get` that includes Nobrainer's losses."""
    objects = {
        "dice": dice,
        "Dice": Dice,
        "focal_tversky": focal_tversky,
        "jaccard": jaccard,
        "Jaccard": Jaccard,
        "tversky": tversky,
        "Tversky": Tversky,
        "elbo": elbo,
        "ELBO": ELBO,
        "wasserstein": wasserstein,
        "Wasserstein": Wasserstein,
        "gradient_penalty": gradient_penalty,
        "GradientPenalty": GradientPenalty,
    }
    with tf.keras.utils.CustomObjectScope(objects):
        return tf.keras.losses.get(loss)
